@@ -11,7 +11,6 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"net/url"
 	"os"
 	"path/filepath"
@@ -230,7 +229,7 @@ func CatFile() (*subprocess.BufferedCmd, error) {
 	return gitNoLFSBuffered("cat-file", "--batch-check")
 }
 
-func DiffIndex(ref string, cached bool, refresh bool) (*bufio.Scanner, error) {
+func DiffIndex(ref string, cached bool, refresh bool, workingDir string) (*bufio.Scanner, error) {
 	if refresh {
 		_, err := gitSimple("update-index", "-q", "--refresh")
 		if err != nil {
@@ -244,6 +243,10 @@ func DiffIndex(ref string, cached bool, refresh bool) (*bufio.Scanner, error) {
 	}
 	args = append(args, ref)
 
+	if workingDir != "" {
+		args = append([]string{"-C", workingDir}, args...)
+	}
+
 	cmd, err := gitBuffered(args...)
 	if err != nil {
 		return nil, err
@@ -531,13 +534,20 @@ func ValidateRemote(remote string) error {
 	if err != nil {
 		return err
 	}
+	return ValidateRemoteFromList(remotes, remote)
+}
+
+// ValidateRemote checks that a named remote is valid for use given a list from
+// RemoteList.  This is completely identical to ValidateRemote, except that it
+// allows caching the remote list.
+func ValidateRemoteFromList(remotes []string, remote string) error {
 	for _, r := range remotes {
 		if r == remote {
 			return nil
 		}
 	}
 
-	if err = ValidateRemoteURL(remote); err == nil {
+	if err := ValidateRemoteURL(remote); err == nil {
 		return nil
 	}
 
@@ -801,6 +811,11 @@ func RootDir() (string, error) {
 	if err != nil {
 		return "", err
 	}
+
+	if len(path) == 0 {
+		return "", errors.New(tr.Tr.Get("no output from `git rev-parse --show-toplevel`"))
+	}
+
 	return tools.CanonicalizePath(path, false)
 }
 
@@ -850,18 +865,24 @@ func GitCommonDir() (string, error) {
 	return tools.CanonicalizePath(path, false)
 }
 
-// GetAllWorkTreeHEADs returns the refs that all worktrees are using as HEADs
+// A git worktree (ref + path)
+type Worktree struct {
+	Ref Ref
+	Dir string
+}
+
+// GetAllWorkTrees returns the refs that all worktrees are using as HEADs plus the worktree's path.
 // This returns all worktrees plus the master working copy, and works even if
 // working dir is actually in a worktree right now
 // Pass in the git storage dir (parent of 'objects') to work from
-func GetAllWorkTreeHEADs(storageDir string) ([]*Ref, error) {
+func GetAllWorkTrees(storageDir string) ([]*Worktree, error) {
 	worktreesdir := filepath.Join(storageDir, "worktrees")
 	dirf, err := os.Open(worktreesdir)
 	if err != nil && !os.IsNotExist(err) {
 		return nil, err
 	}
 
-	var worktrees []*Ref
+	var worktrees []*Worktree
 	if err == nil {
 		// There are some worktrees
 		defer dirf.Close()
@@ -880,29 +901,42 @@ func GetAllWorkTreeHEADs(storageDir string) ([]*Ref, error) {
 					tracerx.Printf("Error reading %v for worktree, skipping: %v", headfile, err)
 					continue
 				}
-				worktrees = append(worktrees, ref)
+
+				// Read the gitdir file to get the location of the git repo
+				dirfile := filepath.Join(worktreesdir, dirfi.Name(), "gitdir")
+				dir, err := parseDirFile(dirfile)
+				if err != nil {
+					tracerx.Printf("Error reading %v for worktree, skipping: %v", dirfile, err)
+					continue
+				}
+
+				worktrees = append(worktrees, &Worktree{*ref, filepath.Dir(dir)})
 			}
 		}
 	}
 
 	// This has only established the separate worktrees, not the original checkout
-	// If the storageDir contains a HEAD file then there is a main checkout
+	// If the storageDir contains a HEAD file and a RootDir then there is a main checkout
 	// as well; this must be resolveable whether you're in the main checkout or
 	// a worktree
 	headfile := filepath.Join(storageDir, "HEAD")
 	ref, err := parseRefFile(headfile)
 	if err == nil {
-		worktrees = append(worktrees, ref)
+		dir, err := RootDir()
+		if err == nil {
+			worktrees = append(worktrees, &Worktree{*ref, dir})
+		} else { // ok if not exists, probably bare repo
+			tracerx.Printf("Error getting toplevel for main checkout, skipping: %v", err)
+		}
 	} else if !os.IsNotExist(err) { // ok if not exists, probably bare repo
 		tracerx.Printf("Error reading %v for main checkout, skipping: %v", headfile, err)
 	}
-
 	return worktrees, nil
 }
 
 // Manually parse a reference file like HEAD and return the Ref it resolves to
 func parseRefFile(filename string) (*Ref, error) {
-	bytes, err := ioutil.ReadFile(filename)
+	bytes, err := os.ReadFile(filename)
 	if err != nil {
 		return nil, err
 	}
@@ -913,6 +947,15 @@ func parseRefFile(filename string) (*Ref, error) {
 	return ResolveRef(contents)
 }
 
+func parseDirFile(filename string) (string, error) {
+	bytes, err := os.ReadFile(filename)
+	if err != nil {
+		return "", err
+	}
+	contents := strings.TrimSpace(string(bytes))
+	return contents, nil
+}
+
 // IsBare returns whether or not a repository is bare. It requires that the
 // current working directory is a repository.
 //
@@ -1313,10 +1356,11 @@ func GetTrackedFiles(pattern string) ([]string, error) {
 
 	var ret []string
 	cmd, err := gitNoLFS(
-		"-c", "core.quotepath=false", // handle special chars in filenames
 		"ls-files",
+		"--ignored",
 		"--cached", // include things which are staged but not committed right now
-		"--",       // no ambiguous patterns
+		"-z",       // handle special chars in filenames
+		"-x",
 		safePattern)
 	if err != nil {
 		return nil, errors.New(tr.Tr.Get("failed to find `git ls-files`: %v", err))
@@ -1328,6 +1372,7 @@ func GetTrackedFiles(pattern string) ([]string, error) {
 	}
 	cmd.Start()
 	scanner := bufio.NewScanner(outp)
+	scanner.Split(tools.SplitOnNul)
 	for scanner.Scan() {
 		line := scanner.Text()
 
